// 
// Arduino UNO Sketch for reading the serial data from the TruePosition GPS board
// and displaying on a 20 characters by 4 lines LCD display with I2C/SPI adapter
// board.
//
// By Benjamin D. Hall, KD5BYB, 4 July 2017
//
// Many special thanks go out to the good folks on the TimeNuts maining list at 
// https://www.febo.com/mailman/listinfo/time-nuts for putting up with all my silly questions and
// for pointing out better ways of doing things.  Especially the use of the time library.
//
// Many thanks go out to the authors of "A Packrat GPS Receiver Project" including
// Gary, WA2OMY and Bruce, WA3YUE.  Without their work in the PDF file here:
// http://www.packratvhf.com/A%20Packrat%20GPS%20Receiver%20Project.pdf
// None of this would have been possible.
//
// Many thanks also go out to JHaskell's Blog:
// <http://jhaskellsblog.blogspot.com/2011/05/serial-comm-fundamentals-on-arduino.html#comment-form>
// Much of the serial parsing code is lifted from this blog post and modified.
//
// NOTES, TERMS OF USE:
//
// By using this code, you understand that I'm NOT a professional programmer and that this is
// my first "real" Arduino project.  It probably has bugs and problems not obviously to me.  I'm
// sure it is so poor that real programmers would look at it and probably lose their last meal.
//
// While it has been working here for several months in various forms with no damage to my hardware,
// your luck may be different.
//
// HARDWARD CONNECTIONS:
//
// This code is targeted towards an Arduino Uno Rev3 development board.
//
// TruePosition is connected to the hardware TX and RX ports of the Arudino Uno via
// a 3.3V to 5V level shifter.  (The Arduino is 5V while the TruePosition is 3.3V Serial TTL)
// Arduino Pin 0 (RX in) is connected via level shifter to TruePosition J10 pin 7
// Arduino Pin 1 (TX out) is connected via level shifter to TruePosition J10 pin 3
// Arduino Pin 3 (1PPS input) is connected directly (no level shifter as it is 5V) to TruePosition J10 pin 11
// Arduino Pin A5 is connected to the LCD display SCL
// Arduino Pin A4 is connected to the LCD Display SDA
//
// REVISION HISTORY
//
// what's new in 10:
// * add output pin switches to allow connection of logic analyzer
// * new function called updateTimeDisplay that only updates the first line of the LCD,
//    saving time and preventing the $STATUS message from being rx'ed improperly
// * new function called updateDisplay that updates only the STATUS portion of the display,
//    and is only called when there is something to update, not every second, becuase
//    status only changes once every five seconds
//
// what's new in 11:
// * start to incorporate menuing
//
// what's new in 12:
// * GPS pages are now written with a for loop
// * updated the main status page to have holdover on forth line
// * added logic to swap pages depending on grounded pins
// * added +1 second offset to fix delta between real time and the display - because the data is 1 second old - yes!
//
// what's new in 13:
// * change from jumper selection of display pages to button push
//
// what's new in 14:
// * remove the debug lines and general clean-up of the code
// * add version and date to initializing message

#include <LiquidCrystal_PCF8574.h>
#include <Time.h>
#include <TimeLib.h>

LiquidCrystal_PCF8574 lcd(0x27);  // set the LCD address to 0x27 for a 20 chars and 4 line display

//receiver status array
// 0=leapsecs, 1=tfom, 2=10M, 3=1PPS, 4=ant, 5=hold, 6=nsat, 7=state, 8=year, 9=month, 10=day, 11=hours, 12=mins, 13=secs, 14=day of week
// 15=survey, 16=numsat
//---> DOP, seconds, and temp have their own variables becuase they are floats, not integers

long stat[] = {9, 9, 9, 9, 9, 9, 9, 99, 9, 9, 9, 9, 9, 9, 9, 9, 9};  
// set the status array to something other than zero
// so we can see when the status actually changes.  State is set to 99 to make sure the clear for values less than 10
// is actually working.

volatile boolean display = true;  //we want to update the time display immediately upon first run

void updateDisplayRequired() {  //this is the interrupt service routine.  It sets display to true when interrupted which then causes the 
  //display update routine to run
  display = true;
}

void setup() {
  lcd.begin(20, 4); // initialize the lcd, 20 columns, 4 rows, LCD is on pins A4 and A5
// initialize serial:
  Serial.begin(9600);  //TruePosition talks 9600 baud
//  mySerial.begin(115200);
  lcd.setBacklight(100);  //turn on backlight
  lcd.home();  //move cursor home
  lcd.clear();  //clear the display
  lcd.setCursor(0, 0);
  lcd.print("INITIALIZING REV 14");  //display startup message and revision
  lcd.setCursor(0, 1);
  lcd.print("4 JULY 2017");  //display date of revision
  pinMode(3, INPUT);  //interrupt input for 1PPS from receiver
  pinMode(11, INPUT_PULLUP);  //the input pin for the pushbutton to switch pages
  delay(2000);
  Serial.println("$PROCEED");  
  delay(2000);
  Serial.println("$PROCEED");
  attachInterrupt(digitalPinToInterrupt(3), updateDisplayRequired, RISING);  //want the 1PPS to trigger the interrupt and update of the display
  lcd.clear();  //clear the display to remove the INITIALIZING message
  lcd.setCursor(0, 0);
}
char startChar = '$'; 
char endChar = '\r';
#define DATABUFFERSIZE 120 // 120 seems to be a good guess
char dataBuffer[DATABUFFERSIZE+1]; //Add 1 for NULL terminator
boolean storeString = false; //This will be our flag to put the data in our buffer
float temp=0.0;  //temp is a float variable
float dop=0.0;  //time dilution of precision is a float variable
boolean newStatus = false; //Setup a variable that will go true only when status is updated so we don't call the display loop so much
int page = 0; //this will determine what menu page we display
int oldpage = 0;  //tracks the previous page displayed to check for chnages
int buttonState = HIGH;
const int buttonPin = 11;  //button is on pin 11
int lastButtonState = LOW;
unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 50; //the debounce time

//
//satellite status array
//rows = channels = 0=ch1, 1=ch2, 2=ch3, 3=ch4, 4=ch5, 5=ch6, 6=ch7, 7=ch8
//columns = 0=prn, 1=el, 2=az, 3=sig
//
long unixdelta = 315964800; //delta between GPS and unix time
// initialize the sats array
int sats[8][4] = {
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0},
{0, 0, 0, 0}
};
char delimiters[] = " $";  //the delimiters used in the TruePosition serial data
char* valPosition;
int channel;

void loop() {

    oldpage = page;  //set the oldpage variable to the page we're currently displaying
    
  // switch routine is from https://www.arduino.cc/en/Tutorial/Debounce
  // but with a modification to only switch the display when the pin goes low
  //
  // read the state of the switch into a local variable:
  int reading = digitalRead(buttonPin);
  // check to see if you just pressed the button
  // (i.e. the input went from HIGH to LOW), and you've waited
  // long enough since the last press to ignore any noise:
  // If the switch changed, due to noise or pressing:
  if (reading != lastButtonState) {
    // reset the debouncing timer
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) > debounceDelay) {
    // whatever the reading is at, it's been there for longer
    // than the debounce delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != buttonState) {
      buttonState = reading;

      // only increment the display page if the button is pressed and is LOW
      if (buttonState == LOW) {
      page = page +1;
      }
      // since we only have display pages 0, 1, and 2, anything greater than 2 should roll over back to I
      if (page > 2) {
        page = 0;
      }
    }
  }

  // Save the reading.  Next time through the loop, it'll be the lastButtonState:
  lastButtonState = reading;
    
    if (page != oldpage){  //we only want to clear the display if the new page isn't the same as the previously displayed page
      lcd.clear();  //clear the display
      lcd.setCursor(0, 0);
      newStatus = true;  //set the newStatus flag, as we've got newStatus to display
    }

    if (page == 0){
      if (display == true)
          {
          updateTimeDisplay();
          display = false;
          }
      if(getSerialString())  //calls getSerialString, returns TRUE if there is new data to parse and we call the parser and probably update the display here
          {
          parser();
          }
      if(newStatus == true)
          {
          updateDisplay();
          newStatus = false;
          }
    }
    if (page == 1){
       if(getSerialString())  //calls getSerialString, returns TRUE if there is new data to parse and we call the parser and probably update the display here
          {
          parser();
          }
      if(newStatus == true)
          {
          gpsPageOne();
          newStatus = false;
          }
    }
  if (page == 2){
       if(getSerialString())  //calls getSerialString, returns TRUE if there is new data to parse and we call the parser and probably update the display here
          {
          parser();
          }
      if(newStatus == true)
          {
          gpsPageTwo();
          newStatus = false;
          }
    }    





}




//end of the loop




//
//GET SERIAL STRING FUNCTION
//
boolean getSerialString(){
    static byte dataBufferIndex = 0;
    while(Serial.available()>0){
        char incomingbyte = Serial.read();
        if(incomingbyte==startChar){
            dataBufferIndex = 0;  //Initialize our dataBufferIndex variable
            storeString = true;
        }
        if(storeString){
            //Let's check our index here, and abort if we're outside our buffer size
            //We use our define here so our buffer size can be easily modified
            //            if(dataBufferIndex==DATABUFFERSIZE){
            //                //Oops, our index is pointing to an array element outside our buffer.
            //                dataBufferIndex = 0;
            //                break;
            //            }
            if(incomingbyte==endChar){
                dataBuffer[dataBufferIndex] = 0; //null terminate the C string
                //Our data string is complete.  return true
                return true;
            }
            else{
                dataBuffer[dataBufferIndex++] = incomingbyte;
                dataBuffer[dataBufferIndex] = 0; //null terminate the C string
            }
        }
        else{
        }
    }

return false;
}
///
/// END OF GET SERIAL STRING FUNCTION
///
///
/// PARSING FUNCTION
///
void parser(){
valPosition = strtok(dataBuffer, delimiters);
long secs = 0;
//receiver status array
// 0=leapsecs, 1=tfom, 2=10M, 3=1PPS, 4=ant, 5=hold, 6=nsat, 7=state, 8=year, 9=month, 10=day, 11=hours, 12=mins, 13=secs, 14=day of week
// 15=survey, 16=numsat
//---> DOP, seconds, and temp have their own variables becuase they are floats, not integers
if (strcmp(valPosition, "CLOCK") == 0)  //have a clock message
{
  valPosition = strtok(NULL, delimiters);   
  secs = atol(valPosition);  //time in seconds
  valPosition = strtok(NULL, delimiters);
  stat[0] = atoi(valPosition);  //leapseconds
  valPosition = strtok(NULL, delimiters);
  stat[1] = atoi(valPosition);  //tfom
  secs = secs - stat[0] + unixdelta + 1;  //subtract out the leap seconds, add in the unix to GPS epoch delta, plus add one second as the message comes out after
  // the 1PPS tick, so we're actually 1 second behind if we don't add this second in.
//
// now let's update the receiver status array
//receiver status array
// 0=leapsecs, 1=tfom, 2=10M, 3=1PPS, 4=ant, 5=hold, 6=nsat, 7=state, 8=year, 9=month, 10=day, 11=hours, 12=mins, 13=secs, 14=day of week
// 15=survey, 16=numsat
//---> DOP, seconds, and temp have their own variables becuase they are floats, not integers
  stat[8] = year(secs);
  stat[9] = month(secs);
  stat[10] = day(secs);
  stat[11] = hour(secs);
  stat[12] = minute(secs);
  stat[13] = second(secs);
  dataBuffer[0] = 0;
}
else if (strcmp(valPosition, "EXTSTATUS") == 0)  //have a extstatus message
// $EXTSTATUS <survey> <numsats> <dop> <rx temperature> <4th unknown int>
// use the float variables temp and dop as the status array is an integer array
{
  valPosition = strtok(NULL, delimiters);  //survey active
  stat[15] = atoi(valPosition);  //survey active
  valPosition = strtok(NULL, delimiters);  //numstat - number of sats used in last fix
  stat[16] = atoi(valPosition);  //numsat
  valPosition = strtok(NULL, delimiters);  //DOP
  dop = atof(valPosition);  //DOP
  valPosition = strtok(NULL, delimiters);  //temp 
  temp = atof(valPosition);  //temperature
//  mySerial.print("temperature: ");
//  mySerial.println (temp);
//receiver status array
// 0=leapsecs, 1=tfom, 2=10M, 3=1PPS, 4=ant, 5=hold, 6=nsat, 7=state, 8=year, 9=month, 10=day, 11=hours, 12=mins, 13=secs, 14=day of week
// 15=survey, 16=numsat
//---> DOP, seconds, and temp have their own variables
  
  
  newStatus = true;
  dataBuffer[0] = 0;
}
else if (strcmp(valPosition, "STATUS") == 0)  //have a status message
// $STATUS <10M> <1PPS> <Ant> <Holdover time> <Nsat> <State>
//receiver status array
// 0=leapsecs, 1=tfom, 2=10M, 3=1PPS, 4=ant, 5=hold, 6=nsat, 7=state, 8=year, 9=month, 10=day, 11=hours, 12=mins, 13=secs, 14=day of week
// 15=survey, 16=numsat
//---> DOP, seconds, and temp have their own variables becuase they are floats, not integers
{
  valPosition = strtok(NULL, delimiters);   
  stat[2] = atoi(valPosition);  //10 MHZ status, 0 = good
  valPosition = strtok(NULL, delimiters);
  stat[3] = atoi(valPosition);  //1pps status, 0 = good
  valPosition = strtok(NULL, delimiters);
  stat[4] = atoi(valPosition);  //ant status, 0 = good
  valPosition = strtok(NULL, delimiters);   
  stat[5] = atoi(valPosition);  //holdover seconds
  valPosition = strtok(NULL, delimiters);
  stat[6] = atoi(valPosition);  //number of sats tracked
  valPosition = strtok(NULL, delimiters);
  stat[7] = atoi(valPosition);  //receiver state
  dataBuffer[0] = 0;
}

else if (strcmp(valPosition, "SAT") == 0)  //have a satellite message
{
// $SAT 4 1 51 103 29
// $SAT <rx channel> <prn> <elevation> <azimuth> <signal>
//satellite status array
//rows = channels = 0=ch1, 1=ch2, 2=ch3, 3=ch4, 4=ch5, 5=ch6, 6=ch7, 7=ch8
//columns = 0=prn, 1=el, 2=az, 3=sig
//int sats[8][4] = {
// ch 0 {ch 0 prn, ch 0 el, 0},
  valPosition = strtok(NULL, delimiters);   
  channel = atoi(valPosition);  //the channel number
  valPosition = strtok(NULL, delimiters);
  sats[channel][0] = atoi(valPosition);  //prn
  valPosition = strtok(NULL, delimiters);
  sats[channel][1] = atoi(valPosition);  //elevation
  valPosition = strtok(NULL, delimiters);   
  sats[channel][2] = atoi(valPosition);  //azimuth
  valPosition = strtok(NULL, delimiters);
  sats[channel][3] = atoi(valPosition);  //signal
  dataBuffer[0] = 0;
}
return;
}
///
/// end of parser function
///

///
/// start of display function
///

void updateTimeDisplay(){
//now let's update the display
//  lcd.clear();
// FIRST LINE:  2017/05/22  12:21:30 (date and time)
  lcd.setCursor(0, 0);
  lcd.print(stat[8]);  //prints the year
  lcd.print("/");  //prints the / mark
//the month may need to have a leading 0 added to it, let's check for that
  if (stat[9] < 10) lcd.print("0"); 
  lcd.print(stat[9]); //month
  lcd.print("/");
//the day may need to have a leading 0 added to it, let's check for that
  if (stat[10] < 10) lcd.print("0");
  lcd.print(stat[10]); //date
//now lets print the time
  lcd.setCursor(12, 0);  //move to x=12, y=0
//the hours may need to have a leading 0 added to it, let's check for that
  if (stat[11] < 10) lcd.print("0");
  lcd.print(stat[11]);
  lcd.print(":");
//the minutes may need to have a leading 0 added to it, let's check for that
  if (stat[12] < 10) lcd.print("0");
  lcd.print(stat[12]);
  lcd.print(":");
//the seconds may need to have a leading 0 added to it, let's check for that
  if (stat[13] < 10) lcd.print("0");
  lcd.print(stat[13]);
}

void updateDisplay(){

//second line:  now lets display sat, 10MHZ status, and 1PPS status
  lcd.setCursor(0, 1);
  lcd.print("SAT=");
  lcd.print(stat[6]);
  lcd.print(" 10MHz=");
  lcd.print(stat[2]);
  lcd.print(" 1PPS=");
  lcd.print(stat[3]);
//third line
  lcd.setCursor(0, 2);
  lcd.print("ANT=");
  lcd.print(stat[4]);
  lcd.print(" LEAP=");
  lcd.print(stat[0]);
  lcd.setCursor(14, 2);
  lcd.print("STA=");
    if (stat[7] < 10){
    lcd.print(" ");
    }
  lcd.print(stat[7]);
  
//forth line
  lcd.setCursor(0, 3);
  lcd.print("TF=");
  lcd.print(stat[1]);
  lcd.setCursor(5, 3);
  lcd.print("H=");
  if (stat[5] < 100){
    lcd.print("0");
  }
  if (stat[5] < 10){
    lcd.print("0");
  }
  lcd.print(stat[5]);
  lcd.setCursor(11, 3);
  lcd.print("TMP=");
  lcd.print(temp);
}

void gpsPageOne(){
  for(int y = 0; y < 4; y++){
    lcd.setCursor(0, y);
    lcd.print(y);
    lcd.print(":P");
    if (sats[y][0] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][0]);  //prn
    lcd.print(" S");
    if (sats[y][3] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][3]);  //signal strength
    lcd.print(" EL");
    if (sats[y][1] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][1]);  //elevation
    lcd.print(" AZ");
    if (sats[y][2] < 10){
      lcd.print(" ");
    }
    if (sats[y][2] < 100){
      lcd.print(" ");
    }
    lcd.print(sats[y][2]);  //azimuth
  }
}

void gpsPageTwo(){
  int x;
  for(int y = 4; y < 8; y++){
    x = y-4;
    lcd.setCursor(0, x);
    lcd.print(y);
    lcd.print(":P");
    if (sats[y][0] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][0]);  //prn
    lcd.print(" S");
    if (sats[y][3] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][3]);  //signal strength
    lcd.print(" EL");
    if (sats[y][1] < 10){
      lcd.print(" ");
    }
    lcd.print(sats[y][1]);  //elevation
    lcd.print(" AZ");
    if (sats[y][2] < 10){
      lcd.print(" ");
    }
    if (sats[y][2] < 100){
      lcd.print(" ");
    }
    lcd.print(sats[y][2]);  //azimuth
  }
}


